#include "main.h"

double       start_time = 0.0;
double       fps        = 30.0;
Ecore_Evas  *ee         = NULL;
Evas        *evas       = NULL;

static void _app_resize(Ecore_Evas *ee);
static int  _app_signal_exit(void *data, int ev_type, void *ev);
static void _app_delete_request(Ecore_Evas *ee);

#define SOFT_X 0
#define FB     1
#define GL_X   2

static Ecore_Event_Handler *_app_handler_exit = NULL;

int
app_start(int argc, const char **argv)
{
   int w = 240;
   int h = 320;
   int rot = 270;
   int im_cache = 4096 * 1024;
   int fn_cache = 512 * 1024;
   int engine = SOFT_X;
   int i;
   
   /* init the app */
   start_time = ecore_time_get();
   ecore_init();
   ecore_app_args_set(argc, argv);
   _app_handler_exit = ecore_event_handler_add(ECORE_EVENT_SIGNAL_EXIT, _app_signal_exit, NULL);
   
   /* create an evas */
   if (!ecore_evas_init()) return -1;
   for (i = 1; i < argc; i++)
     {
	if      ((!strcmp(argv[i], "-x11"))) engine = SOFT_X;
	else if ((!strcmp(argv[i], "-fb")))  engine = FB;
	else if ((!strcmp(argv[i], "-gl")))  engine = GL_X;
	else if ((!strcmp(argv[i], "-g")) && (i < (argc - 1)))
	  {
	     char b1[32], b2[32];
	     
	     i++;
	     b1[0] = 0;
	     b2[0] = 0;
	     sscanf(argv[i], "%30[^x]x%30s", b1, b2);
	     w = atoi(b1);
	     h = atoi(b2);
	     if (((w < 1) || (w > 8000)) ||
		 ((h < 1) || (h > 8000)))
	       {
		  printf("Error parsing options to -g option.\n"
			 "Please see -h help output for syntax.\n");
		  ecore_evas_shutdown();
		  ecore_shutdown();
		  return 0;
	       }
	  }
	else if ((!strcmp(argv[i], "-r")) && (i < (argc - 1)))
	  {
	     i++;
	     rot = atoi(argv[i]);
	     if ((rot != 0) && (rot != 90) && (rot != 270))
	       {
		  printf("Error parsing options to -r option.\n"
			 "Options must be 0, 90 or 270.\n");
		  ecore_evas_shutdown();
		  ecore_shutdown();
		  return 0;
	       }
	  }
	else if ((!strcmp(argv[i], "-ic")) && (i < (argc - 1)))
	  {
	     i++;
	     im_cache = atoi(argv[i]);
	     if ((im_cache < 0) || (im_cache > 65536 * 1024))
	       {
		  printf("Invalid image cache value. use between 0 and 65536\n");
		  ecore_evas_shutdown();
		  ecore_shutdown();
		  return 0;
	       }
	  }
	else if ((!strcmp(argv[i], "-fc")) && (i < (argc - 1)))
	  {
	     i++;
	     fn_cache = atoi(argv[i]);
	     if ((fn_cache < 0) || (fn_cache > 65536 * 1024))
	       {
		  printf("Invalid font cache value. use between 0 and 65536\n");
		  ecore_evas_shutdown();
		  ecore_shutdown();
		  return 0;
	       }
	  }
	else if ((!strcmp(argv[i], "-fr")) && (i < (argc - 1)))
	  {
	     i++;
	     fps = atof(argv[i]);
	     if (fps <= 0.0)
	       {
		  printf("Invalid fps value. must be greater than 0.0\n");
		  ecore_evas_shutdown();
		  ecore_shutdown();
		  return 0;
	       }
	  }
	else if ((!strcmp(argv[i], "-h")))
	  {
	     printf("%s [OPTIONS]\n"
		    "\tWhere OPTIONS is 0 or more of the following:\n"
		    "\n"
		    "-x11       Display using software X11\n"
		    "-fb        Display using the software in the framebuffer\n"
		    "-gl        Display using OpenGL in X11\n"
		    "-g WxH     Set display size to W x H pixels\n"
		    "-r rot     Set rotation of the screen (0, 90, 270)\n"
		    "-ic Kb     Set image cache in Kb\n"
		    "-fc Kb     Set font cache in Kb\n"
		    "-fr fps    Set attempted framerate in frames per second\n"
		    "-h         Display this help\n",
		    argv[0]);
	     ecore_evas_shutdown();
	     ecore_shutdown();
	     return 0;
	  }
     }
   if (engine == SOFT_X)
     {
	ee = ecore_evas_software_x11_new(NULL, 0, 0, 0, w, h);
	if (!ee)
	  ee = ecore_evas_fb_new(NULL, rot, w, h);
     }
   else if (engine == FB)
     {
	ee = ecore_evas_fb_new(NULL, rot, w, h);
	if (!ee)
	  ee = ecore_evas_software_x11_new(NULL, 0, 0, 0, w, h);
     }
   else if (engine == GL_X)
     {
	ee = ecore_evas_gl_x11_new(NULL, 0, 0, 0, w, h);
	if (!ee)
	  ee = ecore_evas_software_x11_new(NULL, 0, 0, 0, w, h);
	if (!ee)
	  ee = ecore_evas_fb_new(NULL, rot,  w, h);
     }
   if (!ee)
     {
	printf("Error. Cannot create canvas. Abort.\n");
	exit(-1);
     }
   ecore_evas_size_min_set(ee, w, h);
   ecore_evas_size_max_set(ee, w, h);
   ecore_evas_callback_delete_request_set(ee, _app_delete_request);
   ecore_evas_callback_resize_set(ee, _app_resize);
   ecore_evas_title_set(ee, "Enlightenment for Embedded");
   ecore_evas_name_class_set(ee, "eem", "main");
//   ecore_evas_avoid_damage_set(ee, 1);
//   ecore_evas_shaped_set(ee, 1);
//   ecore_evas_override_set(ee, 1);
   ecore_evas_show(ee);
   evas = ecore_evas_get(ee);
   evas_image_cache_set(evas, im_cache);
   evas_font_cache_set(evas, fn_cache);
   evas_font_path_append(evas, FN);
   edje_init();
   edje_frametime_set(1.0 / fps);
   return 1;
}

void
app_finish(void)
{
   ecore_evas_free(ee);
   ee = NULL;
   evas = NULL;
   ecore_event_handler_del(_app_handler_exit);
   _app_handler_exit = NULL;
   edje_shutdown();
   ecore_evas_shutdown();
   ecore_shutdown();
}

static void
_app_resize(Ecore_Evas *ee)
{
   Evas_Coord w, h;
   
   evas_output_viewport_get(evas, NULL, NULL, &w, &h);
   bg_resize(w, h);
}

static int
_app_signal_exit(void *data, int ev_type, void *ev)
{
   ecore_main_loop_quit();
   return 1;
}

static void
_app_delete_request(Ecore_Evas *ee)
{
   ecore_main_loop_quit();
}
