#include "main.h"

typedef struct _Smart_Data Smart_Data;
typedef struct _Smart_Item Smart_Item;

struct _Smart_Data
{ 
   Evas_Coord       x, y, w, h;
   Evas_Object     *obj;
   Evas_Object     *clip;
   int              frozen;
   unsigned char    changed : 1;
   unsigned char    horizontal : 1;
   Evas_List       *items;
   struct {
      Evas_Coord    w, h;
   } min, max;
   struct {
      int           num;
      Evas_Coord    offset, cur_offset;
      double        tim, sec;
      Ecore_Timer  *timer;
      void        (*sel_func) (Evas_Object *obj);
      void        (*unsel_func) (Evas_Object *obj);
   } selection;
}; 

struct _Smart_Item
{
   Smart_Data      *sd;
   struct {
      Evas_Coord    w, h;
   } min;
   Evas_Object     *obj;
};

static Smart_Item *_box_smart_adopt(Smart_Data *sd, Evas_Object *obj);
static void        _box_smart_disown(Evas_Object *obj);
static void        _box_smart_reconfigure(Smart_Data *sd);
static void        _box_smart_extents_calcuate(Smart_Data *sd);
static int         _box_cb_timer(void *data);
static void        _box_cb_free(void *data, Evas *e, Evas_Object *obj, void *event_info);

static void _box_smart_init(void);
static void _box_smart_add(Evas_Object *obj);
static void _box_smart_del(Evas_Object *obj);
static void _box_smart_layer_set(Evas_Object *obj, int layer);
static void _box_smart_raise(Evas_Object *obj);
static void _box_smart_lower(Evas_Object *obj);
static void _box_smart_stack_above(Evas_Object *obj, Evas_Object * above);
static void _box_smart_stack_below(Evas_Object *obj, Evas_Object * below);
static void _box_smart_move(Evas_Object *obj, Evas_Coord x, Evas_Coord y);
static void _box_smart_resize(Evas_Object *obj, Evas_Coord w, Evas_Coord h);
static void _box_smart_show(Evas_Object *obj);
static void _box_smart_hide(Evas_Object *obj);
static void _box_smart_color_set(Evas_Object *obj, int r, int g, int b, int a);
static void _box_smart_clip_set(Evas_Object *obj, Evas_Object * clip);
static void _box_smart_clip_unset(Evas_Object *obj);

static Evas_Smart *_smart = NULL;

Evas_Object *
box_add(Evas *evas)
{
   _box_smart_init();
   return evas_object_smart_add(evas, _smart);
}

int
box_freeze(Evas_Object *obj)
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return;
   sd = evas_object_smart_data_get(obj);
   sd->frozen++;
   return sd->frozen;
}

int
box_thaw(Evas_Object *obj)
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return 0;
   sd = evas_object_smart_data_get(obj);
   sd->frozen--;
   if (sd->frozen <= 0) _box_smart_reconfigure(sd);
   return sd->frozen;
}

void
box_orientation_set(Evas_Object *obj, int horizontal)
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return;
   sd = evas_object_smart_data_get(obj);
   if (((sd->horizontal) && (horizontal)) ||
       ((!sd->horizontal) && (!horizontal))) return;
   sd->horizontal = horizontal;
   sd->changed = 1;
   if (sd->frozen <= 0) _box_smart_reconfigure(sd);
}

int
box_pack_start(Evas_Object *obj, Evas_Object *child)
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return 0;
   sd = evas_object_smart_data_get(obj);
   _box_smart_adopt(sd, child);
   sd->items = evas_list_prepend(sd->items, child);
   sd->changed = 1;
   if (sd->frozen <= 0) _box_smart_reconfigure(sd);
   return 0;
}

int
box_pack_end(Evas_Object *obj, Evas_Object *child)
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return 0;
   sd = evas_object_smart_data_get(obj);
   _box_smart_adopt(sd, child);
   sd->items = evas_list_append(sd->items, child);   
   sd->changed = 1;
   if (sd->frozen <= 0) _box_smart_reconfigure(sd);
   return evas_list_count(sd->items) - 1;
}

int
box_pack_count_get(Evas_Object *obj)
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return 0;
   sd = evas_object_smart_data_get(obj);
   return evas_list_count(sd->items);
}

Evas_Object *
box_pack_object_nth(Evas_Object *obj, int n)
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return NULL;
   sd = evas_object_smart_data_get(obj);
   return evas_list_nth(sd->items, n);
}

void
box_unpack(Evas_Object *obj)
{
   Smart_Item *si;
   Smart_Data *sd;
   Evas_List *l;
   int i;
   
   si = evas_object_data_get(obj, "box_data");
   if (!si) return;
   sd = si->sd;
   i = -1;
   for (l = sd->items, i = 0; l; l = l->next, i++)
     {
	if (l->data == obj) break;
     }
   sd->items = evas_list_remove(sd->items, obj);
   _box_smart_disown(obj);
   if (sd->selection.num >= i)
     box_select_set(sd->obj, sd->selection.num - 1);
   sd->changed = 1;
   if (sd->frozen <= 0) _box_smart_reconfigure(sd);
}

void
box_pack_info_set(Evas_Object *obj, Evas_Coord minw, Evas_Coord minh)
{
   Smart_Item *si;
   Smart_Data *sd;
   
   si = evas_object_data_get(obj, "box_data");
   if (!si)
     {
	si = calloc(1, sizeof(Smart_Item));
	if (!si) return;
	si->min.w = 0;
	si->min.h = 0;
	si->obj = obj;
	evas_object_data_set(obj, "box_data", si);
     }
   sd = si->sd;
   if (minw < 0) minw = 0;
   if (minh < 0) minh = 0;
   si->min.w = minw;
   si->min.h = minh;
   if (sd)
     {
	sd->changed = 1;
	if (sd->frozen <= 0) _box_smart_reconfigure(sd);
     }
}

void
box_min_size_get(Evas_Object *obj, Evas_Coord *minw, Evas_Coord *minh)
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return;
   sd = evas_object_smart_data_get(obj);
   if (sd->changed) _box_smart_extents_calcuate(sd);
   if (minw) *minw = sd->min.w;
   if (minh) *minh = sd->min.h;
}

void
box_select_func_set(Evas_Object *obj, void (*sel_func) (Evas_Object *obj), void (*unsel_func) (Evas_Object *obj))
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return;
   sd = evas_object_smart_data_get(obj);
   sd->selection.sel_func = sel_func;
   sd->selection.unsel_func = unsel_func;
}

void
box_select_set(Evas_Object *obj, int n)
{
   Smart_Data *sd;
   Evas_Object *o;
   Smart_Item *si;   
   
   if (strcmp(evas_object_type_get(obj), "box")) return;
   sd = evas_object_smart_data_get(obj);
   si = NULL;
   o = evas_list_nth(sd->items, sd->selection.num);
   if (o) si = evas_object_data_get(o, "box_data");
   if (si)
     {
	if (sd->selection.unsel_func) sd->selection.unsel_func(si->obj);
     }
   sd->selection.num = n;
   si = NULL;
   o = evas_list_nth(sd->items, sd->selection.num);
   if (o) si = evas_object_data_get(o, "box_data");
   if (si)
     {
	if (sd->selection.sel_func) sd->selection.sel_func(si->obj);
     }
   sd->selection.tim = ecore_time_get();
   sd->selection.offset = sd->selection.cur_offset;
   sd->changed = 1;
   if (sd->frozen <= 0) _box_smart_reconfigure(sd);
   if ((!sd->selection.timer) && (sd->selection.sec > 0.0))
     {
	sd->selection.timer = ecore_timer_add(1.0 / fps, _box_cb_timer, obj);
     }
}

int
box_select_get(Evas_Object *obj)
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return 0;
   sd = evas_object_smart_data_get(obj);
   return sd->selection.num;
}

void
box_select_time_set(Evas_Object *obj, double sec)
{
   Smart_Data *sd;
   
   if (strcmp(evas_object_type_get(obj), "box")) return;
   sd = evas_object_smart_data_get(obj);
   sd->selection.sec = sec;
   sd->changed = 1;
   if (sd->frozen <= 0) _box_smart_reconfigure(sd);
}

double
box_select_time_get(Evas_Object *obj)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (strcmp(evas_object_type_get(obj), "box")) return 0.0;
   return sd->selection.tim;
}

static Smart_Item *
_box_smart_adopt(Smart_Data *sd, Evas_Object *obj)
{
   Smart_Item *si;
   
   si = evas_object_data_get(obj, "box_data");
   if (!si)
     {
	si = calloc(1, sizeof(Smart_Item));
	if (!si) return NULL;
	si->min.w = 0;
	si->min.h = 0;
	si->obj = obj;
	evas_object_data_set(obj, "box_data", si);
     }
   si->sd = sd;
   evas_object_smart_member_add(obj, si->sd->obj);
   evas_object_clip_set(obj, si->sd->clip);
   evas_object_stack_below(obj, si->sd->clip);
   evas_object_event_callback_add(obj, EVAS_CALLBACK_FREE, _box_cb_free, NULL);
   return si;
}

static void
_box_smart_disown(Evas_Object *obj)
{
   Smart_Item *si;
   
   si = evas_object_data_get(obj, "box_data");
   if (!si) return;
   evas_object_smart_member_del(obj);
   evas_object_clip_unset(obj);
   evas_object_data_del(obj, "box_data");
   evas_object_event_callback_del(obj, EVAS_CALLBACK_FREE, _box_cb_free);
   free(si);
}

static void
_box_smart_reconfigure(Smart_Data *sd)
{
   Evas_Coord x, y, w, h, xx, yy, off, offsz;
   Evas_List *l;
   int minw, minh;
   int count, i;
   double t;
   double v;

   if (!sd->changed) return;

   t = ecore_time_get();
   
   x = sd->x;
   y = sd->y;
   w = sd->w;
   h = sd->h;

   _box_smart_extents_calcuate(sd);
   minw = sd->min.w;
   minh = sd->min.h;
   count = evas_list_count(sd->items);
   if (sd->horizontal)
     {
	if (h < minh) h = minh;	
     }
   else
     {
	if (w < minw) w = minw;
     }
   
   xx = 0;
   yy = 0;
   off = 0;
   offsz = 0;
   for (i = 0, l = sd->items; l; l = l->next, i++)
     {
	Smart_Item *si;
	Evas_Object *obj;
	
	obj = l->data;
	si = evas_object_data_get(obj, "box_data");
	if (sd->horizontal)
	  {
	     Evas_Coord ww;
	     
	     ww = si->min.w;
	     if (i == sd->selection.num)
	       {
		  off = xx;
		  offsz = ww;
		  break;
	       }
	     xx += ww;
	  }
	else
	  {
	     Evas_Coord hh;
	     
	     hh = si->min.h;
	     if (i == sd->selection.num)
	       {
		  off = yy;
		  offsz = hh;
		  break;
	       }
	     yy += hh;
	  }
     }
   
   if (sd->horizontal)
     {
	if (off < sd->selection.cur_offset)
	  off = off;
	else if ((off + offsz) > (sd->selection.cur_offset + w))
	  off = off - w + offsz;
	else 
	  off = sd->selection.cur_offset;
     }
   else
     {
	if (off < sd->selection.cur_offset)
	  off = off;
	else if ((off + offsz) > (sd->selection.cur_offset + h))
	  off = off - h + offsz;
	else 
	  off = sd->selection.cur_offset;
     }
   t -= sd->selection.tim;
   
   if (t < sd->selection.sec)
     {
	v = 1.0 - sin((t / sd->selection.sec) * M_PI / 2.0);   
	v = 1.0 - (v * v);
	
	off = (off * v) + (sd->selection.offset * (1.0 - v));
     }
   sd->selection.cur_offset = off;
   
   if (sd->horizontal)
     {
	xx = x - (int)sd->selection.cur_offset;
	yy = y;
     }
   else
     {
	xx = x;
	yy = y - (int)sd->selection.cur_offset;
     }
   for (l = sd->items; l; l = l->next)
     {
	Smart_Item *si;
	Evas_Object *obj;
	
	obj = l->data;
	si = evas_object_data_get(obj, "box_data");
	if (sd->horizontal)
	  {
	     Evas_Coord ww, hh;
	     
	     ww = si->min.w;
	     hh = h;
	     evas_object_move(obj, xx, yy);
	     evas_object_resize(obj, ww, hh);
	     xx += ww;
	  }
	else
	  {
	     Evas_Coord ww, hh;
	     
	     ww = w;
	     hh = si->min.h;
	     evas_object_move(obj, xx, yy);
	     evas_object_resize(obj, ww, hh);
	     yy += hh;
	  }
     }
   evas_object_move(sd->clip, -9999999, -9999999);
   evas_object_resize(sd->clip, 99999999, 99999999);
   sd->changed = 0;
}

/* FIXME: need to have min/max size calc routines */
static void
_box_smart_extents_calcuate(Smart_Data *sd)
{
   Evas_List *l;
   int minw, minh;
   
   sd->max.w = -1; /* max < 0 == unlimited */
   sd->max.h = -1;
   
   minw = 0;
   minh = 0;
   for (l = sd->items; l; l = l->next)
     {
	Smart_Item *si;
	Evas_Object *obj;
	
	obj = l->data;
	si = evas_object_data_get(obj, "box_data");	
	if (sd->horizontal)
	  {
	     if (minh < si->min.h) minh = si->min.h;
	     minw += si->min.w;
	  }
	else
	  {
	     if (minw < si->min.w) minw = si->min.w;
	     minh += si->min.h;
	  }
     }
   sd->min.w = minw;
   sd->min.h = minh;
}

static int
_box_cb_timer(void *data)
{
   Smart_Data *sd;
   double t;
   
   sd = evas_object_smart_data_get((Evas_Object *)data);
   
   t = ecore_time_get() - sd->selection.tim;
   
   sd->changed = 1;
   _box_smart_reconfigure(sd);
   
   if (t < sd->selection.sec) return 1;
   sd->selection.timer = NULL;
   return 0;
}

static void
_box_cb_free(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
   box_unpack(obj);
}

static void
_box_smart_init(void)
{
   if (_smart) return;
   _smart = evas_smart_new("box",
			     _box_smart_add,
			     _box_smart_del,
			     _box_smart_layer_set,
			     _box_smart_raise,
			     _box_smart_lower,
			     _box_smart_stack_above,
			     _box_smart_stack_below,
			     _box_smart_move,
			     _box_smart_resize,
			     _box_smart_show,
			     _box_smart_hide,
			     _box_smart_color_set,
			     _box_smart_clip_set,
			     _box_smart_clip_unset,
			     NULL);
}

static void
_box_smart_add(Evas_Object *obj)
{
   Smart_Data *sd;
   
   sd = calloc(1, sizeof(Smart_Data));
   if (!sd) return;
   sd->obj = obj;
   sd->x = 0;
   sd->y = 0;
   sd->w = 0;
   sd->h = 0;
   sd->clip = evas_object_rectangle_add(evas_object_evas_get(obj));
   evas_object_smart_member_add(sd->clip, obj);
   evas_object_color_set(sd->clip, 255, 255, 255, 255);
   evas_object_smart_data_set(obj, sd);
}
   
static void
_box_smart_del(Evas_Object *obj)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
   evas_object_del(sd->clip);
   if (sd->selection.timer) ecore_timer_del(sd->selection.timer);
   while (sd->items)
     {
	Evas_Object *o;
	
	o = sd->items->data;
	sd->items = evas_list_remove(sd->items, o);
	_box_smart_disown(o);
     }
   free(sd);
}
   
static void
_box_smart_layer_set(Evas_Object *obj, int layer)
{
   Smart_Data *sd;
      
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
     {
	Evas_List *l;
	
	for (l = sd->items; l; l = l->next)
	  {
	     Evas_Object *o;
	     
	     o = l->data;
	     evas_object_layer_set(o, layer);
	  }
     }
}

static void
_box_smart_raise(Evas_Object *obj)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
     {
	Evas_List *l;
	
	for (l = sd->items; l; l = l->next)
	  {
	     Evas_Object *o;
	     
	     o = l->data;
	     evas_object_raise(o);
	  }
	evas_object_raise(sd->clip);
     }
}

static void
_box_smart_lower(Evas_Object *obj)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (!sd) return; 
     {
	Evas_List *l;
	
	evas_object_lower(sd->clip);
	for (l = evas_list_last(sd->items); l; l = l->prev)
	  {
	     Evas_Object *o;
	     
	     o = l->data;
	     evas_object_lower(o);
	  }
     }
}
                                                             
static void
_box_smart_stack_above(Evas_Object *obj, Evas_Object *above)
{
   Smart_Data *sd;

   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
     {
	Evas_List *l;
	
	evas_object_stack_above(sd->clip, above);
	for (l = evas_list_last(sd->items); l; l = l->prev)
	  {
	     Evas_Object *o;
	     
	     o = l->data;
	     evas_object_stack_above(o, above);
	  }
     }
}
   
static void
_box_smart_stack_below(Evas_Object *obj, Evas_Object *below)
{
   Smart_Data *sd;
      
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
     {
	Evas_List *l;
	
	for (l = sd->items; l; l = l->next)
	  {
	     Evas_Object *o;
	     
	     o = l->data;
	     evas_object_stack_below(o, below);
	  }
	evas_object_stack_below(sd->clip, below);
     }
}

static void
_box_smart_move(Evas_Object *obj, Evas_Coord x, Evas_Coord y)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
   if ((x == sd->x) && (y == sd->y)) return;
   sd->x = x;
   sd->y = y;
   sd->changed = 1;
   _box_smart_reconfigure(sd);
}

static void
_box_smart_resize(Evas_Object *obj, Evas_Coord w, Evas_Coord h)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
   if ((w == sd->w) && (h == sd->h)) return;
   sd->w = w;
   sd->h = h;
   sd->changed = 1;
   _box_smart_reconfigure(sd);
}

static void
_box_smart_show(Evas_Object *obj)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
   evas_object_show(sd->clip);
}

static void
_box_smart_hide(Evas_Object *obj)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
   evas_object_hide(sd->clip);
}

static void
_box_smart_color_set(Evas_Object *obj, int r, int g, int b, int a)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
   evas_object_color_set(sd->clip, r, g, b, a);
}

static void
_box_smart_clip_set(Evas_Object *obj, Evas_Object * clip)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
   evas_object_clip_set(sd->clip, clip);
}

static void
_box_smart_clip_unset(Evas_Object *obj)
{
   Smart_Data *sd;
   
   sd = evas_object_smart_data_get(obj);
   if (!sd) return;
   evas_object_clip_unset(sd->clip);
}  
